Attribute VB_Name = "DirectionalFunctions"
'directional functions
' by K. SATO, 2004/11/13
' Division of Earth and Planetary Sciences, Kyoto University

Option Explicit
Const pi = 3.14159265358979, d2r = pi / 180, r2d = 180 / pi

'001
Function P_separate_bearing(str)
  'use separate_num_str
  Dim s_flag, numeric_flag As Boolean
  Dim c, first_character, str0 As String
  Dim strike, dip As Single
  Dim ew, nsew As String
  Dim result As Variant
  
  str = LCase(Trim(str))
  first_character = Mid(str, 1, 1)
  
  s_flag = False
  Select Case first_character
    Case "n"
      str0 = Mid(str, 2)
    Case "s"
      str0 = Mid(str, 2)
      s_flag = True
    Case Else
      str0 = str
  End Select
  
  result = separate_num_str(str0)
  strike = result(1)
  ew = result(2)
  dip = result(3)
  nsew = result(4)
  
  If s_flag Then
    Select Case ew
      Case "e"
        ew = "w"
      Case "w"
        ew = "e"
    End Select
  End If
  
  P_separate_bearing = Array(strike, ew, dip, nsew)
End Function

'002
Function P_bering2azpl(strike, ew, dip, nsew)
  Dim az, pl As Single
  
  ew = LCase(Trim(ew))
  nsew = LCase(Trim(nsew))
  
  Select Case ew
  Case "e"
    Select Case nsew
    Case "n", "w"
      az = 270 + strike
    Case "s", "e"
      az = 90 + strike
    Case Else
      az = "error in nsew"
    End Select
  Case "w"
    Select Case nsew
    Case "n", "e"
      az = 90 - strike
    Case "s", "w"
      az = 270 - strike
    Case Else
      az = "error in nsew"
    End Select
  Case Else
      az = "error in ew"
  End Select
  
  pl = dip
  P_bering2azpl = Array(az, pl)
End Function

'003
Function separate_num_str(str)
  Dim result(1 To 100)
  Dim c As String
  Dim numeric_flag0, numeric_flag1 As Boolean
  Dim k, i As Integer
  
  k = 1
  c = Mid(str, 1, 1)
  result(k) = c
  numeric_flag0 = IsNumeric(c)
  
  For i = 2 To Len(str)
    c = Mid(str, i, 1)
    If (c = ".") And (numeric_flag0 = True) Then
      numeric_flag1 = True
    Else
      numeric_flag1 = IsNumeric(c)
    End If
    
    If numeric_flag0 = numeric_flag1 Then
      result(k) = result(k) & c
    Else
      k = k + 1
      If k > 100 Then
        result(100) = "error: over 100 elements"
        Exit For
      End If
      result(k) = c
      numeric_flag0 = numeric_flag1
    End If
  Next
  
  separate_num_str = result
End Function

'004
Function fault_untilting(Paz, Ppl, Laz, Lpl, sense, note, Baz, Bpl, _
                         Optional dor = 100, _
                         Optional Baz0 = 0, Optional Bpl0 = 0, _
                         Optional original_to_be_horizontal_flag As Boolean = False)
  'use azpl2uvector, cross_prod, dyad, mat_vec_mult, dot_prod
  Dim n, s, b As Variant                      'vector(1 to 3), 1:s, 2:b, 3:n
  Dim bn, bn0, rot_axis As Variant            'vector(1 to 3)
  Dim rot_angle, lon, col As Single
  Dim projector(1 To 3, 1 To 3) As Variant    'matrix(1 to 3, 1 to 3)
  Dim nn As Variant                           'matrix(1 to 3, 1 to 3)
  Dim unit_matrix3(1 To 3, 1 To 3) As Variant
  Dim Q0(1 To 3, 1 To 3), Q1(1 To 3, 1 To 3), Q2(1 To 3, 1 To 3), Q3, Qb0 As Variant
  Dim Q, Qax, QaxT As Variant                    'matrix(1 to 3, 1 to 3)
  Dim i, j As Integer
  Dim n_r, s_r, b_r, bn_r As Variant
  Dim s_r_downward(1 To 3) As Single
  Dim azpl As Variant
  Dim dip_vector As Variant
  'result
  Dim Paz_r, Ppl_r, Laz_r, Lpl_r As Variant
  Dim sense_r, note_r As String
  Dim Baz_r, Bpl_r As Variant
  Dim err As Variant
  Dim rake As Variant '0 < rake < 90 deg
  
  Paz_r = ""
  Ppl_r = ""
  Laz_r = ""
  Lpl_r = ""
  sense_r = ""
  note_r = ""
  Baz_r = ""
  Bpl_r = ""
  err = ""
  rake = ""
  If IsEmpty(Paz) Or IsEmpty(Ppl) Or IsEmpty(Laz) Or IsEmpty(Lpl) Or IsEmpty(sense) Or _
    IsEmpty(Baz) Or IsEmpty(Bpl) Then
    note_r = "insufficient arguments"
    fault_untilting = Array(Paz_r, Ppl_r, Laz_r, Lpl_r, sense_r, note_r, Baz_r, Bpl_r, err, rake)
    Exit Function
  End If
  
  If IsEmpty(dor) Then dor = 100
  If IsEmpty(Baz0) Then Baz0 = 0
  If IsEmpty(Bpl0) Then Bpl0 = 0
  
  unit_matrix3(1, 1) = 1
  unit_matrix3(1, 2) = 0
  unit_matrix3(1, 3) = 0
  unit_matrix3(2, 1) = 0
  unit_matrix3(2, 2) = 1
  unit_matrix3(2, 3) = 0
  unit_matrix3(3, 1) = 0
  unit_matrix3(3, 2) = 0
  unit_matrix3(3, 3) = 1

  n = azpl2uvector(Paz, Ppl - 90)      'upward fault normal
  s = azpl2uvector(Laz, Lpl)           'downward slip orientation
  
  'projection of slip vector onto fault plane
  err = Abs(90 - Application.Acos(dot_prod(s, n)) * r2d)
  nn = dyad(n, n)
  For i = 1 To 3
    For j = 1 To 3
      projector(i, j) = unit_matrix3(i, j) - nn(i, j)
    Next
  Next
  s = mat_vec_mult(projector, s)
  s = normalize_vector(s)
  
  b = cross_prod(n, s)
  
  'shear sense correction
  sense = LCase(sense)
  Select Case sense
    Case "n"
    Case "r"
      For i = 1 To 3
        s(i) = -s(i)
        b(i) = -b(i)
      Next
    Case "s"
      If b(3) < 0 Then
        For i = 1 To 3
          s(i) = -s(i)
          b(i) = -b(i)
        Next
      End If
    Case "d"
      If b(3) > 0 Then
        For i = 1 To 3
          s(i) = -s(i)
          b(i) = -b(i)
        Next
      End If
    Case Else
      err = "error in input sense"
  End Select
  
  'bedding plane
  bn = azpl2uvector(Baz, Bpl - 90)
  bn0 = azpl2uvector(Baz0, Bpl0 - 90)

  'rotation axis and angle
  rot_axis = cross_prod(bn, bn0)
  rot_axis = normalize_vector(rot_axis)
  lon = Application.Atan2(rot_axis(1), rot_axis(2))
  col = Application.Acos(rot_axis(3))
  
  rot_angle = Application.Acos(dot_prod(bn, bn0)) * dor / 100
  
  'rotation tensor
  Q0(1, 1) = Cos(rot_angle)
  Q0(1, 2) = -Sin(rot_angle)
  Q0(1, 3) = 0
  Q0(2, 1) = Sin(rot_angle)
  Q0(2, 2) = Cos(rot_angle)
  Q0(2, 3) = 0
  Q0(3, 1) = 0
  Q0(3, 2) = 0
  Q0(3, 3) = 1
  
  Q1(1, 1) = Cos(col)
  Q1(1, 2) = 0
  Q1(1, 3) = Sin(col)
  Q1(2, 1) = 0
  Q1(2, 2) = 1
  Q1(2, 3) = 0
  Q1(3, 1) = -Sin(col)
  Q1(3, 2) = 0
  Q1(3, 3) = Cos(col)
  
  Q2(1, 1) = Cos(lon)
  Q2(1, 2) = -Sin(lon)
  Q2(1, 3) = 0
  Q2(2, 1) = Sin(lon)
  Q2(2, 2) = Cos(lon)
  Q2(2, 3) = 0
  Q2(3, 1) = 0
  Q2(3, 2) = 0
  Q2(3, 3) = 1
  
  Qax = Application.MMult(Q2, Q1)
  QaxT = Application.Transpose(Qax)
  Q = Application.MMult(Application.MMult(Qax, Q0), QaxT)
  
  'rotate assumed original bedding to be horizontal
  If original_to_be_horizontal_flag Then
    Q1(1, 1) = Cos(Baz0 * d2r)
    Q1(1, 2) = -Sin(Baz0 * d2r)
    Q1(1, 3) = 0
    Q1(2, 1) = Sin(Baz0 * d2r)
    Q1(2, 2) = Cos(Baz0 * d2r)
    Q1(2, 3) = 0
    Q1(3, 1) = 0
    Q1(3, 2) = 0
    Q1(3, 3) = 1
    
    Q2(1, 1) = 1
    Q2(1, 2) = 0
    Q2(1, 3) = 0
    Q2(2, 1) = 0
    Q2(2, 2) = Cos(Bpl0 * d2r)
    Q2(2, 3) = -Sin(Bpl0 * d2r)
    Q2(3, 1) = 0
    Q2(3, 2) = Sin(Bpl0 * d2r)
    Q2(3, 3) = Cos(Bpl0 * d2r)
    
    Q3 = Application.Transpose(Q1)
    
    Qb0 = Application.MMult(Application.MMult(Q3, Q2), Q1)
    
    Q = Application.MMult(Qb0, Q)
'    Q = Application.MMult(Q1, Q)
  End If
    
  'rotation
  n_r = mat_vec_mult(Q, n)
  s_r = mat_vec_mult(Q, s)
  b_r = cross_prod(n_r, s_r)
  bn_r = mat_vec_mult(Q, bn)
  
  'check hanging-wall and foot wall
  If n_r(3) < 0 Then
    For i = 1 To 3
      s_r(i) = -s_r(i)
      b_r(i) = -b_r(i)
      n_r(i) = -n_r(i)
    Next
  End If
  
  'shear sense
  If s_r(3) > 0 Then
    sense_r = "r"
  ElseIf s_r(3) < 0 Then
    sense_r = "n"
  Else
    If b_r(3) > 0 Then
      sense_r = "s"
    ElseIf b_r(3) < 0 Then
      sense_r = "d"
    Else
      sense_r = "horizontal fault"
    End If
  End If
  
  'azpl
  azpl = uvector2azpl(n_r)
  Paz_r = azpl(1)
  Ppl_r = azpl(2) + 90
  
  If s_r(3) < 0 Then
    For i = 1 To 3
      s_r_downward(i) = s_r(i)
    Next
  Else
    For i = 1 To 3
      s_r_downward(i) = -s_r(i)
    Next
  End If
  azpl = uvector2azpl(s_r_downward)
  Laz_r = azpl(1)
  Lpl_r = azpl(2)
  
  azpl = uvector2azpl(bn_r)
  Baz_r = azpl(1)
  Bpl_r = azpl(2) + 90
  
  dip_vector = azpl2uvector(Paz_r, Ppl_r)
  rake = Application.Acos(dot_prod(dip_vector, s_r))
  rake = Abs(rake - pi / 2) * r2d
  
  note_r = note & "[untilting=" & str(dor) & _
                "%][assumed_original_bedding=" & str(Baz0) & "/" & str(Bpl0) & _
                 "][rotation_to_horizontal_original=" & original_to_be_horizontal_flag & "]"
  
  fault_untilting = Array(Paz_r, Ppl_r, Laz_r, Lpl_r, sense_r, note_r, Baz_r, Bpl_r, err, rake)
End Function

'005
Function azpl2uvector(az, pl)
  'az, pl in [deg]
  Dim lon, col As Single
  Dim v(1 To 3)
  
  lon = (90 - az) * d2r
  col = (90 + pl) * d2r
  v(1) = Sin(col) * Cos(lon)
  v(2) = Sin(col) * Sin(lon)
  v(3) = Cos(col)
  azpl2uvector = v
End Function

'006
Function cross_prod(v1, v2)
'Function cross_prod(v1 As Range, v2 As Range)
  Dim v(1 To 3)
  
  v(1) = v1(2) * v2(3) - v1(3) * v2(2)
  v(2) = v1(3) * v2(1) - v1(1) * v2(3)
  v(3) = v1(1) * v2(2) - v1(2) * v2(1)
  cross_prod = v
End Function

'007
Function dyad(v1, v2)
  Dim d(1 To 3, 1 To 3) As Variant
  Dim i, j As Integer
  
  For i = 1 To 3
    For j = 1 To 3
      d(i, j) = v1(i) * v2(j)
    Next
  Next
  dyad = d
End Function

'008
Function mat_vec_mult(A, x)
  Dim y(1 To 3) As Variant
  Dim i, j As Integer
  
  For i = 1 To 3
    y(i) = 0
    For j = 1 To 3
      y(i) = y(i) + A(i, j) * x(j)
    Next
  Next
  mat_vec_mult = y
End Function

'009
Function dot_prod(v1, v2)
  Dim dot As Single
  Dim i As Integer
  
  dot = 0
  For i = 1 To 3
    dot = dot + v1(i) * v2(i)
  Next
  dot_prod = dot
End Function

'010
Function normalize_vector(v)
  Dim vn(1 To 3) As Single
  Dim l As Single
  Dim i As Integer
    
  l = 0
  For i = 1 To 3
    l = l + v(i) * v(i)
  Next
  l = Sqr(l)
  For i = 1 To 3
    vn(i) = v(i) / l
  Next
  normalize_vector = vn
End Function

'011
Function uvector2azpl(v)
  'az, pl in [deg]
  Dim lon, col As Single
  Dim azpl(1 To 2) As Single
  
  lon = Application.Atan2(v(1), v(2))
  col = Application.Acos(v(3))
  
  azpl(1) = (90 - lon * r2d + 360)
  azpl(1) = azpl(1) - (azpl(1) \ 360) * 360
  azpl(2) = col * r2d - 90
  uvector2azpl = azpl
End Function

